#
# Generic network interface plugin for Indigo 5
#
# Copyright 2011-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import socket
import asyn
import asyn.resolve
import cyin.devstate
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *


#
# Map line ending configuration strings to actual string values
#
EOL = {
#	symbolic: (text, regex pattern)
	"nl":	('\n',		r'\n'),
	"cr":	('\r',		r'(?s)\r'),
	"crnl":	('\r\n',	r'\r?\n'),
	"nul":	('\0',		r'(?s)\0'),
	"none":	('',		r'')
}


#
# A Live TCP connection.
# Used by both outgoing and incoming connection devices.
#
class _TCPConnection(cyin.devstate.Device):
	""" A live established TCP connection.

		This class manages a single TCP connection once established, until
		it is terminated by us or the peer. Subclasses implement different
		ways to establish the connection. The object is reusable.

		As long as a TCPConnection is connected, its asyn scanner will look for
		matches in incoming data. Any "Recognized" event can specify a regular
		expression that is automatically looked for; upon a match, the event is
		triggered. If a full incoming line has no match, any Unrecognized events
		are triggered instead. Either way, data is consumed line-by-line.

		Data can be written with an explicit action.
	"""
	config_version = 2	# has port state

	port = cyin.PluginProperty(eval=True, check=[check_port])
	eol = cyin.PluginProperty()
	lastinput = cyin.PluginProperty(required=False, reconfigure=False)
	lastinput_var = cyin.NamedVariable(name="lastinput")

	active_port = cyin.DeviceState(name="port", type=int)
	incount = cyin.DeviceState(type=int)
	outcount = cyin.DeviceState(type=int)

	def start_common(self, address, flags):
		""" Common initialization of _TCPConnection objects. This is not a constructor. """
		super(_TCPConnection, self).start()
		self.stream = None			# active connection object
		self.active_port = 0
		self.reset_action = None
		try:
			self.resolved = socket.getaddrinfo(address or None, self.port,
				0, socket.SOCK_STREAM, 0, socket.AI_CANONNAME | flags)
			return True
		except socket.error as e:
			self.fail_hard(e.args[1])
			self.resolved = None
			return False

	def stop(self):
		self._disconnect()
		super(_TCPConnection, self).stop()

	def start_stream(self, socket):
		""" Given a TCP socket, set up a live connection. """
		self.active_port = socket.getsockname()[1]
		self.connector = None
		self.incount = 0
		self.outcount = 0
		self.proceed("connected", recovered=True)
		self.stream = cyin.plugin.stream(socket, callout=self._event)
		self.update_scan()

	@action
	def disconnect(self, action):
		""" Forcibly disconnect the TCP Connection right now. """
		self._disconnect()

	def _disconnect(self, changed=False):
		""" Return the connection to the idle state, with extreme prejudice. """
		if self.stream:
			self.stream.close()
			self.stream = None
			self.active_port = 0
			changed = True
		if changed:
			self.proceed("disconnected")
			if self.ready():
				cyin.plugin.schedule(self.reload) # follow-on actions

	@action
	def sendText(self, action):
		""" Send a line of text. """
		if self.state != "connected":
			return error(self.name, "not connected - ignoring text output")
		if action.partial:
			self.stream.write_a(action.text)
		else:
			# action.text may contain arbitrary escaped bytes.
			# Use the "latin-1" coding to transcribe them as-is.
			text = action.text + EOL[self.eol][0]
			self.stream.write(text.encode("latin-1"))
			self.outcount += 1

	def update_scan(self):
		""" Rebuild the connection scan object based on active triggers. """
		if self.state == "connected":
			rules = []
			_, eols = EOL[self.eol]
			for event in Recognized.all_matching(self):
				rules.append(('(' + event.scan_rule() + ')' + eols, event))
			debug(self.name, "configured for", len(rules), "input event(s)")
			if eols:
				unrecognized = r'(.*?)' + eols		# eat up to eol string
			else:
				unrecognized = r'(.+?)'				# eat what you can
			rules.append((unrecognized, '_UNRECOGNIZED'))
			self.stream.scan = asyn.scan.Regex(rules)

	def _event(self, ctx, *args):
		""" Handle upstream live-connection events. """
		if ctx.error:
			return error(self.name, 'error', ctx)
		if ctx.state == 'END':
			return self._disconnect()
		elif ctx.scan:
			matches = ctx.match.groups()
			if self.lastinput:
				self.lastinput_var = matches[0]
			if ctx.state == '_UNRECOGNIZED':
				debug(self.name, "unrecognized input", repr(matches[0]))
				Unrecognized.trigger(self)		# fire all unrecognized-data events
			else:
				event = ctx.state
				debug(self.name, "recognized", event.name, repr(matches[0]))
				if event.lastinput:
					event.lastinput_var = matches[1] if len(matches) > 1 else matches[0]
				event.trigger_me()				# fire just the trigger that matched
			self.incount += 1


#
# An outgoing (connecting) TCP port
#
class Outgoing(_TCPConnection):
	""" An outgoing TCP connection.

		The connection can be triggered with an explicit action.
		The device can also be set to "persist", in which case a connection
		is continuously attempted as soon as the previous one drops.
	"""
	address = cyin.PluginProperty(required=False, eval=True, check=[check_host])
	persist = cyin.PluginProperty(type=bool)

	def start(self):
		""" Set up an outgoing TCP object. """
		self.connector = None
		if self.start_common(self.address, 0):
			self.proceed("idle")
			if self.persist:
				self.setup()

	@action
	def connect(self, action):
		""" Initiate an outbound connection to the configured address. """
		if self.state == "connected":
			error(self.name, "already connected - ignoring connect request")
		elif self.state == "connecting":
			error(self.name, "already attempting connection - ignoring connect request")
		else:
			self.setup()

	def setup(self, ctx=None):
		""" Initiate an outbound connection effort. """
		self.proceed("connecting", log="connecting to %s" % (self.address or "localhost"))
		self.connector = asyn.resolve.TCPConnector(cyin.plugin, self.resolved, callout=self._connected)

	def _connected(self, ctx, socket=None):
		""" Connection attempt complete (one way or another). """
		if ctx.error:
			if asyn.resolve.transient_error(ctx.error):
				self.fail_soft(ctx.error)
			else:
				self.fail_hard(ctx.error)
			return
		if ctx.state == 'CANCELLED':	# we're going to die
			return
		self.connector = None	# done with it
		self.start_stream(socket)

	def _disconnect(self):
		""" Disconnect with prejudice. """
		changed = False
		if self.connector:
			self.connector.close()
			self.connector = None
			changed = True
		return _TCPConnection._disconnect(self, changed)

	def reload(self, ctx):
		""" Follow-on actions after disconnect. """
		assert self.state == "disconnected"
		if self.persist:
			debug(self.name, "will re-establish its persistent connection")
			cyin.plugin.schedule(lambda ctx: self.setup(), after=0.05)
		else:
			self.proceed("idle")

	def display_address(self):
		return "%s:%s" % ((self.address or "localhost"), self.port)


#
# An incoming (listening) TCP port
#
class Incoming(_TCPConnection):
	""" An incoming TCP connection.

		We always listen for incoming connections to our port.
		Only one connection at a time is accepted; the kernel
		will generally hold a few additional attempts for a while.
		This setup is not designed to deal with simultaneous
		connection attempts.
	"""
	def start(self):
		""" Initialize an incoming TCP object. """
		self.listener = None
		if self.start_common(None, socket.AI_PASSIVE):
			self.listener = cyin.plugin.listener(self.resolved, callout=self._accept)
			self._listen()

	def stop(self):
		if self.listener:
			self.listener.close()
			self.listener = None
		super(Incoming, self).stop()

	def _listen(self):
		""" Turn on listening for incoming connections. """
		self.proceed("listening", log="waiting for incoming connection")
		self.listener.accepting(True)

	def _accept(self, ctx, socket=None):
		""" Incoming connection accepted or failed. """
		if ctx.error:
			return self.fail_hard(ctx)
		if ctx.state == "accept":
			self.listener.accepting(False)	# one at a time
			self.start_stream(socket)

	def reload(self, ctx):
		""" Follow-on actions after disconnect. """
		assert self.state == "disconnected"
		self._listen()

	def display_address(self):
		return "port %s" % self.port


#
# Events for TCP devices.
#
class _TCPEvent(cyin.Event):
	""" Common features of TCP device events. """

	device = cyin.PluginProperty(type=cyin.device)

	def matches(self, dev):
		return self.device == dev


class Recognized(_TCPEvent):
	""" A Recognized event fires when a particular (regex) string is found in an input line. """
	pattern = cyin.PluginProperty(eval=True)
	strict = cyin.PluginProperty(type=bool)
	lastinput = cyin.PluginProperty(required=False, reconfigure=False)
	lastinput_var = cyin.NamedVariable(name="lastinput")

	def scan_rule(self):
		""" Deliver an asyn.Regex scan rule for this event. """
		if self.strict:
			return self.pattern
		else:
			return r'.*' + self.pattern + '.*'

	# reconfigure our device scanner whenever an event for it comes or goes
	def start(self):
		if self.device:
			self.device.update_scan()

	def stop(self):
		if self.device:
			self.device.update_scan()
		super(Recognized, self).stop()


class Unrecognized(_TCPEvent):
	""" A (whole) line was received on a TCP device that no Recognized event matched. """
	pass


#
# Action classes
#
class Connect(cyin.Action): pass

class Disconnect(cyin.Action): pass

def escape_string(s):
	return (s
		.encode('latin1')			# make bytes
		.decode('unicode-escape')	# decode escapes and turn it back into a str
	)

class SendText(cyin.DeviceAction):
	text = cyin.PluginProperty("text", escape_string, dynamic_type=str, eval=True)
	partial = cyin.PluginProperty("partial", bool)

	class UI(cyin.ConfigUI):
		def description(self):
			return 'Send%s "%s" to %s' % (
				" partial" if self.partial else "", repr(self.text), self.device.name)


#
# The plugin.
# There's literally nothing going on in here. (Behold the power of cyin!)
#
class Plugin(cyin.asynplugin.Plugin):
	pass
